import functools
import logging
import time
from functools import partial

from tornado.ioloop import IOLoop

from end2end import metric
from end2end.connectors import Connector


def run_once(name):
    def _tmp(func):
        def _call(self_, *args, **kwargs):
            if not hasattr(self_, '_calls'):
                self_._calls = {}
            if name not in self_._calls:
                self_._calls[name] = func(self_, *args, **kwargs)
            return self_._calls[name]

        return _call

    return _tmp


def _log_with_warning(message, timeout):
    if timeout:
        logging.warning('Timed out for {}'.format(message))
    else:
        logging.info(message)


class DataToSend(object):
    def __init__(self, value: int, connector: Connector):
        self.value = value
        self.connector = connector
        self.start_time = time.time()

    @run_once('data_sent')
    def on_data_sent(self, timeout=False):
        _log_with_warning('Send callback for {}, {}'.format(self.connector.name, self.value), timeout)
        self.connector.send_metric.on_value(time.time() - self.start_time)

    @run_once('async_received')
    def on_async_received(self, timeout=False):
        _log_with_warning('Async callback for {}, {}'.format(self.connector.name, self.value), timeout)
        self.connector.async_metric.on_value(time.time() - self.start_time)

    @run_once('async_max_received')
    def on_async_max_received(self, timeout=False):
        _log_with_warning('Async max callback for {}, {}'.format(self.connector.name, self.value), timeout)
        self.connector.async_max_metric.on_value(time.time() - self.start_time)

    @run_once('sync_received')
    def on_sync_received(self, timeout=False):
        _log_with_warning('Sync callback for {}, {}'.format(self.connector.name, self.value), timeout)
        self.connector.sync_metric.on_value(time.time() - self.start_time)

    def on_timeout_passed(self, sync_used):
        self.on_data_sent(True)
        self.on_async_received(True)
        self.on_async_max_received(True)
        if sync_used:
            self.on_sync_received(True)


class _Registry(object):
    def __init__(self):
        self._connectors = []
        self.value = int(time.time())
        self.rps = metric.instance().create_call_counter('RPS')

    def items(self):
        return tuple(self._connectors)

    def set_items(self, connectors):
        for c in self._connectors:
            c.deinitialize()
        self._connectors.clear()
        for c in connectors:
            self._connectors.append(c)
            c.initialize(partial(self._register_invocation, c))

    def _on_connector_called(self):
        pass

    def _register_invocation(self, connector):
        use_sync_calculator = connector.interval >= 2.

        def _invoke():
            if not connector.active:
                return
            self.value += 1

            data = DataToSend(self.value, connector)

            connector.send_and_receive(data, use_sync_calculator)
            IOLoop.instance().call_later(connector.interval, _invoke)
            IOLoop.instance().call_later(connector.max_wait,
                                         functools.partial(data.on_timeout_passed, use_sync_calculator))
            self.rps.on_call()

        IOLoop.instance().add_callback(_invoke)


__REGISTRY = _Registry()


def instance():
    return __REGISTRY



import logging

import yaml
import click
from tornado.ioloop import IOLoop

from end2end.connectors import registry
from end2end.connectors.factory import load_connectors
from end2end.server import start_http_server
from end2end import security


@click.command()
@click.option('--config', help='Configuration file name')
@click.option('--port', help='Port to listen on')
@click.option('--token', help='Token to use. By default berry token is used')
def start(config, port, token):
    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s\t%(levelname)s\t%(message)s')
    logging.getLogger('tornado.curl_httpclient').setLevel(logging.WARN)
    logging.info('Reading configuration from {}'.format(config))
    with open(config, 'r') as f:
        items = yaml.load(f)
        connectors = load_connectors(items.get('connectors', {}))
    if not connectors:
        raise Exception('No connectors information found in {}'.format(config))
    if token:
        security.use_static_token(token)
    else:
        security.use_berry_token('end2end_nakadi')
    start_http_server(port)
    registry.instance().set_items(connectors)
    IOLoop.instance().start()


if __name__ == '__main__':
    start()



import math

import time
import os


class EMA(object):
    def __init__(self, rpm, mins):
        self.alfa = 1. - math.exp(-5. / (mins * rpm))
        self.value = None

    def add(self, value):
        if self.value is None:
            self.value = value
        else:
            self.value = value * self.alfa + (1 - self.alfa) * self.value


class Percentile(object):
    def __init__(self, sample_size):
        self.sample_size = sample_size
        self.sample = []
        self.p95 = None
        self.p98 = None
        self.p99 = None

    def add(self, value):
        self.sample.append(value)
        size = len(self.sample)
        if size == self.sample_size:
            sorted_sample = sorted(self.sample);
            self.p95 = sorted_sample[int(math.ceil((size * 95) / 100)) - 1]
            self.p98 = sorted_sample[int(math.ceil((size * 98) / 100)) - 1]
            self.p99 = sorted_sample[int(math.ceil((size * 99) / 100)) - 1]
            del self.sample[0]


class Named(object):
    def __init__(self, name):
        self.name = name


class Metric(Named):
    def __init__(self, name, rpm):
        super(Metric, self).__init__(name)
        self.count = 0
        self.last = 0
        self.emas = {str(i): EMA(rpm, i) for i in (1, 5, 15)}
        self.percentile = Percentile(os.getenv('PERCENTILE_SAMPLE_SIZE', 1000))

    def on_value(self, secs):
        [ema.add(secs) for ema in self.emas.values()]
        self.last = secs
        self.count += 1
        self.percentile.add(secs)

    def dump(self):
        r = {'m{}'.format(k): v.value for k, v in self.emas.items()}
        r['count'] = self.count
        r['last'] = self.last
        r['p95'] = self.percentile.p95
        r['p98'] = self.percentile.p98
        r['p99'] = self.percentile.p99
        return r

    def __str__(self):
        return ','.join('{}:{}'.format(k, v.value) for k, v in self.dump().items())


class CallCounter(Metric):
    def __init__(self, name):
        super(CallCounter, self).__init__(name, 60)
        self.last_update = int(time.time())
        self._counter = 0

    def _check_time(self):
        for i in range(0, int(time.time() - self.last_update)):
            self.on_value(self._counter)
            self._counter = 0
            self.last_update += 1

    def on_call(self):
        self._check_time()
        self._counter += 1


class StatusCounter(Named):
    def __init__(self, name):
        super().__init__(name)
        self.counts = {}

    def on_new_status(self, status):
        status = str(status)
        if status not in self.counts:
            self.counts[status] = 1
        else:
            self.counts[status] += 1

    def dump(self):
        return {'status_{}'.format(k): v for k, v in self.counts.items()}


class MetricsRegistry(object):
    def __init__(self):
        self._metrics = {}

    def _register(self, metric):
        self._metrics[metric.name] = metric
        return metric

    def create_metric(self, name, rpm):
        return self._register(Metric(name, rpm))

    def create_status_counter(self, name):
        return self._register(StatusCounter(name))

    def create_call_counter(self, name):
        return self._register(CallCounter(name))

    def delete(self, metric):
        if metric.name in self._metrics and self._metrics[metric.name] == metric:
            del self._metrics[metric.name]

    def dump(self):
        result = {}
        for k, v in self._metrics.items():
            x = result
            for i in k.split('.'):
                if i not in x:
                    x[i] = {}
                x = x[i]
            x.update(v.dump())
        return result


__REGISTRY = MetricsRegistry()


def instance():
    return __REGISTRY



import logging
import tokens

__TOKEN_PROVIDER = None


def get_token():
    if not __TOKEN_PROVIDER:
        return None
    return __TOKEN_PROVIDER()


def set_provider(provider):
    global __TOKEN_PROVIDER
    __TOKEN_PROVIDER = provider


def use_static_token(token):
    set_provider(lambda: token)


def use_berry_token(app_name):
    tokens.configure()
    tokens.manage(
        app_name,
        ['nakadi.event_stream.read', 'nakadi.event_stream.write', 'nakadi.event_type.write', 'uid'])
    tokens.start()

    def _get_token():
        try:
            return tokens.get(app_name)
        except Exception as e:
            logging.error('Failed to get token for {}'.format(app_name), exc_info=e)
            return ''

    set_provider(_get_token)



from end2end import metric


class Connector(object):
    def __init__(self, name, **kwargs):
        self.config = kwargs
        self.name = name
        self.interval = float(kwargs.get('interval', 10))
        self.max_wait = float(kwargs.get('max_wait', 60))
        if self.interval <= 0:
            self.interval = 1
        self.sync_metric = metric.instance().create_metric('connector.{}.sync'.format(name), 60. / self.interval)
        self.async_metric = metric.instance().create_metric('connector.{}.async'.format(name), 60. / self.interval)
        self.async_max_metric = metric.instance().create_metric('connector.{}.async_max'.format(name),
                                                                60. / self.interval)
        self.send_metric = metric.instance().create_metric('connector.{}.send'.format(name), 60. / self.interval)
        self.send_rpm = metric.instance().create_call_counter('connector.{}.rps'.format(name))
        self.active = True

    def send_and_receive(self, data, use_sync):
        self.send_rpm.on_call()

    def deinitialize(self):
        metric.instance().delete(self.sync_metric)
        metric.instance().delete(self.async_metric)
        metric.instance().delete(self.async_max_metric)
        metric.instance().delete(self.send_metric)
        metric.instance().delete(self.send_rpm)
        self.active = False



import json
import logging
import random
import string
import threading
import uuid
from contextlib import closing
from datetime import datetime, tzinfo, timedelta

import requests
from tornado.curl_httpclient import CurlAsyncHTTPClient
from tornado.ioloop import IOLoop

from end2end import metric
from end2end.connectors import Connector
from end2end.connectors.registry import DataToSend
from end2end.security import get_token

STREAM_SEPARATOR = '\n'.encode('UTF-8')

READ_TIMEOUT = 20

CONNECT_TIMEOUT = 1

ZERO_DELTA = timedelta(0)


class UtcTimeZone(tzinfo):
    def utcoffset(self, dt):
        return ZERO_DELTA

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return ZERO_DELTA


UTC_INSTANCE = UtcTimeZone()


def _create_event_type_description(topic):
    return {
        'name': topic,
        'owning_application': 'end2end_monitor',
        'category': 'business',
        'enrichment_strategies': ['metadata_enrichment'],
        'schema': {
            'type': 'json_schema',
            'schema': json.dumps({
                'properties': {
                    'value': {
                        'type': 'number',
                        'description': 'Value that is used for checking'
                    },
                    'instance_id': {
                        'type': 'string',
                        'description': 'Type describing instance id of this e2e checker'
                    },
                    'trash': {
                        'type': 'string',
                        'description': 'Trash to test message load'
                    },
                },
                'required': ['value', 'instance_id', 'trash'],
                'title': 'Schema for end2end monitoring',
            })
        }
    }


def stream_events(base_url, topic_name, verify, cursors_, instance_id, value_callback):
    streaming_cursors = json.loads(json.dumps(cursors_))
    thread_def = {
        'stopped': False
    }

    def _stream_internal():
        while not thread_def['stopped']:
            try:
                headers = {
                    'Accept': 'application/json',
                    'X-nakadi-cursors': json.dumps(streaming_cursors),
                    'Authorization': 'Bearer {}'.format(get_token())
                }
                headers.update(_generate_authorization())
                response = requests.get(
                    '{}/event-types/{}/events?batch_limit=1'.format(base_url, topic_name),
                    headers=headers,
                    verify=verify,
                    timeout=(1, 40),
                    stream=True)
                with closing(response) as r:
                    if r.status_code == 200:
                        for line in r.iter_lines(chunk_size=1):
                            batch = json.loads(line.decode('utf-8'))
                            update_cursors(streaming_cursors, batch['cursor'])
                            for evt in [x for x in batch.get('events', []) if x['instance_id'] == instance_id]:
                                value_callback(evt['value'])
                            if thread_def['stopped']:
                                logging.info('Breaking because thread for {} stopped'.format(topic_name))
                    else:
                        logging.error('Streaming for {} returned status code {}'.format(topic_name, r.status_code))
            except Exception as e:
                logging.error("Failed to process events", exc_info=e)

    t = threading.Thread(target=_stream_internal)
    t.setDaemon(True)
    t.start()
    return thread_def


def update_cursors(cursors_list, cursor):
    [c.update(cursor) for c in cursors_list if c['partition'] == cursor['partition']]


def _generate_authorization():
    return {'Authorization': 'Bearer {}'.format(get_token())}


def _prepare_defaults(params):
    h = params.get('headers', {})
    h.update(_generate_authorization())
    h['Accept-Encoding'] = 'gzip;q=0,deflate'
    params['headers'] = h
    if 'connect_timeout' not in params:
        params['connect_timeout'] = 5
    if 'request_timeout' not in params:
        params['request_timeout'] = 5


class RT(object):
    def __init__(self, base_url, max_clients, verify=True):
        self.http_client = CurlAsyncHTTPClient(IOLoop.instance(), max_clients=max_clients)
        self.base_url = base_url
        self.verify = verify

    def fetch(self, url, callback, **kwargs):
        kwargs['request_timeout'] = 60
        _prepare_defaults(kwargs)
        kwargs['validate_cert'] = self.verify
        return self.http_client.fetch('{}{}'.format(self.base_url, url), callback=callback, **kwargs)

    def stream(self, url, cb, complete_cb, **kwargs):
        kwargs['request_timeout'] = 60
        kwargs['validate_cert'] = self.verify
        _prepare_defaults(kwargs)

        return self.http_client.fetch(
            '{}{}'.format(self.base_url, url),
            callback=complete_cb,
            streaming_callback=cb,
            **kwargs)


def _generate_trash(trash_size):
    return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(trash_size))


class NakadiConnector(Connector):
    def __init__(self, name, **kwargs):
        super(NakadiConnector, self).__init__(name, **kwargs)
        self.topic = kwargs['topic']
        self.receivers = int(kwargs.get('receivers', 1))
        self.initialized_receivers = [None for i in range(0, self.receivers)]
        self.r = RT(kwargs['host'], int(5 + 5 / self.interval), kwargs['verify'])
        self.instance_id = str(uuid.uuid4())
        self.trash = _generate_trash(kwargs['trash-size'])
        self.async_callbacks = {}
        self.sync_callbacks = {}
        self.init_callback = None
        self.cursors = None
        self.guard = threading.Condition()
        self.status_counter = metric.instance().create_status_counter('connector.{}.publish'.format(self.name))

    def deinitialize(self):
        if self.init_callback is not None:
            return IOLoop.instance().call_later(1, self.deinitialize)
        for t in self.initialized_receivers:
            t['stopped'] = True
        metric.instance().delete(self.status_counter)
        super(NakadiConnector, self).deinitialize()

    def initialize(self, init_callback):
        self.init_callback = init_callback

        def _ensure_event_type_exists():
            logging.debug('Checking if event type {} exists'.format(self.topic))

            def _on_event_type(r):
                if r.code == 404:
                    return _create_event_type()
                elif r.code == 200:
                    return _prepare_cursors()
                else:
                    logging.error('Failed to check for event type ({} {}), retrying'.format(r.code, r.body))
                    return _ensure_event_type_exists()

            return self.r.fetch('/event-types/{}'.format(self.topic), _on_event_type, method='GET')

        def _create_event_type():
            logging.debug('Creating event type {}'.format(self.topic))

            def _on_event_type_created(r):
                if r.code == 201:
                    logging.info('Created event type {}'.format(self.topic))
                    return _prepare_cursors()
                else:
                    logging.error('Failed to create event type {}, code: {}, message: {}, retrying'.format(
                        self.topic, r.code, r.body))
                    return _create_event_type()

            self.r.fetch(
                '/event-types',
                _on_event_type_created,
                method='POST',
                headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
                body=json.dumps(_create_event_type_description(self.topic))
            )

        def _prepare_cursors():
            def _on_cursors_fetched(r):
                if r.code == 200:
                    self.cursors = [{
                                        'partition': p['partition'],
                                        'offset': p['newest_available_offset']
                                    } for p in json.loads(r.body.decode('UTF-8'))]

                    self.start_streaming()
                    self.init_callback = None
                    return init_callback()
                else:
                    logging.error('Failed to read partitions info for {}. Status code: {}, content: {}'.format(
                        self.topic, r.code, r.body))
                    return _prepare_cursors()

            return self.r.fetch(
                '/event-types/{}/partitions'.format(self.topic),
                _on_cursors_fetched,
                method='GET',
                headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
            )

        _ensure_event_type_exists()

    def start_streaming(self):
        for i in range(0, self.receivers):
            self.initialized_receivers[i] = stream_events(
                self.r.base_url, self.topic, self.r.verify, self.cursors, self.instance_id, self.value_callback)

    def send_and_receive(self, data: DataToSend, use_sync: bool):
        super(NakadiConnector, self).send_and_receive(data, use_sync)
        self.register_async_callback(data.value, data.on_async_received, data.on_async_max_received)
        if use_sync:
            self.sync_callbacks[data.value] = data.on_sync_received

        def _on_event_pushed(r):
            self.status_counter.on_new_status(r.code)
            if r.code == 200:
                logging.info('successfully published event')
                data.on_data_sent()
                if use_sync:
                    return self._receive(data.value)
            else:
                logging.error('Failed to publish event to {}, status code: {}, content: {}'.format(
                    self.topic, r.code, r.body))
                self.delete_async_callback(data.value)

        self.r.fetch(
            '/event-types/{}/events'.format(self.topic),
            _on_event_pushed,
            method='POST',
            headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
            body=json.dumps([{
                'metadata': {
                    'eid': str(uuid.uuid4()),
                    'event_type': self.topic,
                    'occurred_at': datetime.now().replace(tzinfo=UTC_INSTANCE).isoformat()
                },
                'value': data.value,
                'instance_id': self.instance_id,
                'trash': self.trash}]).encode('utf-8')
        )

    def _receive(self, value):
        attempts_left = [5]

        def _on_response(r):
            if r.code != 200:
                logging.warning('status {} and body {} while fetching for value {}'.format(r.code, r.body, value))
                return _fetch_again()
            batch = json.loads(r.body.decode('UTF-8'))
            update_cursors(self.cursors, batch['cursor'])
            for e in [x for x in batch['events'] if x['instance_id'] == self.instance_id]:
                if e['value'] in self.sync_callbacks:
                    self.sync_callbacks.pop(e['value'])()
            if value in self.sync_callbacks:
                return _fetch_again()

        def _fetch_again():
            if attempts_left[0] == 0:
                return logging.error('no attempts left to fetch value {}'.format(value))
            attempts_left[0] -= 1
            return self.r.fetch(
                '/event-types/{}/events?batch_limit=1&stream_limit=1'.format(self.topic),
                _on_response,
                request_timeout=20,
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json',
                    'X-nakadi-cursors': json.dumps(self.cursors)
                }
            )

        _fetch_again()

    def register_async_callback(self, value, async_callback, async_max_callback):
        if self.receivers:
            with self.guard:
                self.async_callbacks[value] = self.receivers, async_callback, async_max_callback

    def delete_async_callback(self, value):
        if self.receivers:
            with self.guard:
                if value in self.async_callbacks:
                    del self.async_callbacks[value]

    def value_callback(self, value):
        async_max_callback = None
        async_callback = None
        with self.guard:
            if value in self.async_callbacks:
                count, async_callback, async_max_callback = self.async_callbacks.pop(value)
                if count != self.receivers:
                    async_callback = None
                count -= 1
                if count > 0:
                    self.async_callbacks[value] = count, None, async_max_callback
                    async_max_callback = None
            else:
                logging.error('Callback for instance {} and value {} is not found'.format(self.instance_id, value))
        if async_callback:
            IOLoop.instance().add_callback(async_callback)
        if async_max_callback:
            IOLoop.instance().add_callback(async_max_callback)



from end2end.connectors.nakadi import NakadiConnector

CONNECTORS = {
    'nakadi': NakadiConnector
}


def __load_connector(name, spec):
    type_ = spec.get('type')
    if type_ not in CONNECTORS:
        raise Exception('Connector type {} is not supported. Supported types are: {}'.format(type_, CONNECTORS.keys()))

    return CONNECTORS[type_](name, **spec)


def load_connectors(json):
    return [__load_connector(x, y) for x, y in json.items()]



import json

from tornado.httpserver import HTTPServer
from tornado.web import RequestHandler, Application, url

from end2end import metric
from end2end.connectors import registry
from end2end.connectors.factory import load_connectors


class MetricsHandler(RequestHandler):
    def get(self):
        return self.write(metric.instance().dump())


class HealthHandler(RequestHandler):
    def get(self):
        return self.write('OK')


class ConnectorsHandler(RequestHandler):
    def get(self):
        return self.write({c.name: c.config for c in registry.instance().items()})

    def post(self):
        registry.instance().set_items(load_connectors(json.loads(self.request.body.decode('utf-8'))))
        return self.write('OK')


def start_http_server(port):
    application = Application([
        url(r'/health', HealthHandler),
        url(r'/metrics', MetricsHandler),
        url(r'/connectors', ConnectorsHandler)
    ])
    HTTPServer(application).listen(port)
